#!/usr/bin/env python
# -*- coding: utf-8 -*
from multiprocessing import Process, Lock, Pipe
import multiprocessing, argparse, time, sys, os, signal, traceback, socket, sched, atexit
import paho.mqtt.client as mqtt  # pip install paho-mqtt
import hvac
import logging
logging.basicConfig(filename='/var/log/hvac-mqtt-bridge.log',format='%(asctime)s %(levelname)s: %(message)s',level=logging.INFO)

dirname = os.path.dirname(os.path.abspath(__file__)) + '/'
CONFIG = os.getenv('HVACMQTTCONFIG', '/etc/hvac-mqtt-bridge/hvac-mqtt-bridge.conf')

class Config(object):
    def __init__(self, filename=CONFIG):
        self.config = {}
        execfile(filename, self.config)

    def get(self, key, default='special empty value'):
        v = self.config.get(key, default)
        if v == 'special empty value':
            logging.error("Configuration parameter '%s' should be specified" % key)
            sys.exit(2)
        return v

def unhandeledException(e):
    logging.error("Unhandeled exception reseived" + e)
#    trace = open('/tmp/transfer-unhandled-exception.log', 'w+')
#    traceback.print_exc(None, trace)
#    trace.close()

class ReadDevice(Process):
    def __init__(self, pipe, divicemac, device, conf):
        super(ReadDevice, self).__init__()
        self.pipe         = pipe
        self.divicemac    = divicemac
        self.device       = device
        self.conf         = conf


    def run(self):
#	logging.info("PID child " + %os.getpid())
#        print('PID child %d' % os.getpid())
        mqttc= mqtt.Client(self.conf.get('mqtt_clientid', 'hvac')+'-%s-%s' % (self.divicemac,os.getpid()), clean_session=self.conf.get('mqtt_clean_session', False))
        mqttc.will_set(self.conf.get('mqtt_topic_prefix', '/hvac'), payload="Disconnect", qos=self.conf.get('mqtt_qos', 0), retain=False)
        mqttc.reconnect_delay_set(min_delay=3, max_delay=30)
        mqttc.username_pw_set(self.conf.get('mqtt_username'), self.conf.get('mqtt_password'))
        mqttc.connect(self.conf.get('mqtt_broker', 'localhost'), int(self.conf.get('mqtt_port', '1883')), 60)
        mqttc.loop_start()
        try:
            if self.device.auth():
                self.run = True
#                print self.device.type
            else:
                self.run = False
            while self.run:
                try:
                    if self.pipe.poll(self.conf.get('loop_time', 60)):
                        result = self.pipe.recv()
                        if type(result) == tuple and len(result) == 2:
                            (cmd, opts) = result
#                            print ( result)
                        if cmd=='set_temp' and 16<=int(opts)<=31:
                        	self.device.set_temp(int(opts))
                        elif cmd=='set_power' and 0<=int(opts[0])<=1:
                                self.device.set_power(int(opts[0]))
                        elif cmd=='set_ac_mode' and 0<=int(opts[0])<=4:
                                self.device.set_ac_mode(int(opts[0]))
                        elif cmd=='set_fan_speed' and 0<=int(opts[0])<=8:
                                self.device.set_fan_speed(int(opts[0]))
                        elif cmd=='set_quiet' and 0<=int(opts[0])<=1:
                                self.device.set_quiet(int(opts[0]))
                        elif cmd=='set_fan_direction' and 0<=int(opts[0])<=6:
                                self.device.set_fan_direction(int(opts[0]))
                        elif cmd=='set_stepless_max' and 0<=int(opts[0])<=100:
                                self.device.set_stepless_max(int(opts[0]))
                        elif cmd=='set_light' and 0<=int(opts[0])<=1:
                                self.device.set_light(int(opts[0]))
                        elif cmd=='set_health' and 0<=int(opts[0])<=1:
                                self.device.set_health(int(opts[0]))
                        elif cmd=='set_sleep' and 0<=int(opts[0])<=1:
                                self.device.set_sleep(int(opts[0]))
                        elif cmd=='set_energy_save' and 0<=int(opts[0])<=1:
                                self.device.set_energy_save(int(opts[0]))
                        elif cmd=='set_timer':
				(opts1, opts2, opts3) = opts.split(',')
				if 0<=int(opts1)<=23 and 0<=int(opts2)<=59 and 0<=int(opts3)<=2:
				    self.device.set_timer(int(opts1), int(opts2), int(opts3))
                        else:
                            if result == 'STOP':
                                self.shutdown()
                                mqttc.loop_stop()
                                return
                    else:
                        try:
                            data = self.device.get_full_status()
                        except socket.timeout:
                            mqttc.loop_stop()
                            return
                        except Exception, e:
                            unhandeledException(e)
                            mqttc.loop_stop()
                            return
                        for key in data:
                            if type(data[key]).__name__ == 'list':
                                pass
                            else:
                                mqttc.publish('%s/%s/%s'%(self.conf.get('mqtt_topic_prefix', '/hvac'), self.divicemac, key), data[key], qos=self.conf.get('mqtt_qos', 0), retain=self.conf.get('mqtt_retain', False))
                except Exception, e:
                    unhandeledException(e)
                    mqttc.loop_stop()
                    return
            mqttc.loop_stop()
            return

        except KeyboardInterrupt:
            mqttc.loop_stop()
            return

        except Exception, e:
            unhandeledException(e)
            mqttc.loop_stop()
            return

def main():
    try:
        conf = Config()
    except Exception, e:
        logging.error("Cannot load configuration from file %s: %s" % (CONFIG, str(e)))
#        print "Cannot load configuration from file %s: %s" % (CONFIG, str(e))
        sys.exit(2)

    jobs = []
    pipes = []
    founddevices = {}

    def on_message(client, pipes, msg):
        cmd = msg.topic.split('/')
        devicemac = cmd[2]
        command = cmd[4]
        if cmd[3] == 'cmd':
            try:
                for (ID, pipe) in pipes:
                    if ID==devicemac:
#                        print 'send to pipe %s' % ID
                        pipe.send((command, msg.payload))
            except:
                logging.error("Unexpected error:", sys.exc_info()[0])
                print "Unexpected error:", sys.exc_info()[0]
                raise

    def on_disconnect(client, empty, rc):
        logging.warning("Disconnect, reason: " + str(rc))
        logging.warning("Disconnect, reason: " + str(client))
#        print("Disconnect, reason: " + str(rc))
#        print("Disconnect, reason: " + str(client))
        time.sleep(10)

    def on_kill(mqttc, jobs):
        mqttc.loop_stop()
        for j in jobs:
            j.join()

    def on_connect(client, userdata, flags, rc):
        client.publish(conf.get('mqtt_topic_prefix', '/hvac'), 'Connect')

    def on_log(mosq, obj, level, string):
        print(string)

    mqttc = mqtt.Client(conf.get('mqtt_clientid', 'hvac')+'-%s'%os.getpid(), clean_session=conf.get('mqtt_clean_session', False), userdata=pipes)
    mqttc.on_message = on_message
    mqttc.on_connect = on_connect
    mqttc.on_disconnect = on_disconnect
    #mqttc.on_log = on_log
    mqttc.will_set(conf.get('mqtt_topic_prefix', '/hvac'), payload="Disconnect", qos=conf.get('mqtt_qos', 0), retain=False)
    mqttc.reconnect_delay_set(min_delay=3, max_delay=30)
    mqttc.username_pw_set(conf.get('mqtt_username'), conf.get('mqtt_password'))
    mqttc.connect(conf.get('mqtt_broker', 'localhost'), int(conf.get('mqtt_port', '1883')), 60)
    mqttc.subscribe(conf.get('mqtt_topic_prefix', '/hvac') + '/+/cmd/#', qos=conf.get('mqtt_qos', 0))
    mqttc.loop_start()

    atexit.register(on_kill, mqttc, jobs)

    run = True
    while run:
        try:
            for idx, j in enumerate(jobs):
                if not j.is_alive():
                    try:
                        j.join()
                    except:
                        pass
                    try:
                        del founddevices[j.pid]
                    except:
                        pass
                    try:
                        del jobs[idx]
                    except:
                        pass
            logging.info("hvac discover")
#            print "hvac discover"
            devices = hvac.discover(timeout=conf.get('lookup_timeout', 5))
            for device in devices:
                divicemac = ''.join(format(x, '02x') for x in device.mac)
                if divicemac not in founddevices.values():
                    logging.info("found: {} {}".format(device.host[0], ''.join(format(x, '02x') for x in device.mac)))
#                    print "found: {} {}".format(device.host[0], ''.join(format(x, '02x') for x in device.mac))

                    transportPipe, MasterPipe = Pipe()
                    pipes.append((divicemac, MasterPipe))

#                    print "found: {} {}".format(device.host[0], ''.join(format(x, '02x') for x in device.mac))
                    p = ReadDevice(transportPipe, divicemac, device, conf)
                    jobs.append(p)
                    p.start()
                    founddevices[p.pid] = divicemac
                    time.sleep(2)
            mqttc.user_data_set(pipes)
            time.sleep(conf.get('rediscover_time', 600))
        except KeyboardInterrupt:
            run = False
        except Exception, e:
            run = False
            unhandeledException(e)
        except SignalException_SIGKILL:
            run = False

    mqttc.loop_stop()
    for j in jobs:
        j.join()
    return

main()
